#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <math.h>
#include "uart.h"
#include "dpy.h"
#include "gps.h"
#include "vibrate.h"


#define KNOTS_TO_KMPH 1.852


int get_checksum(char *data)
// Gets the XOR Checksum (of every byte) of a sentence.
// For NMEA, this is the last byte before the end of the char.
{
	uint8_t len=0;
	uint16_t checksum=0;
	printf("#TGPS1.0 * Data is %s\r\n", data); /* Print out the line */
	len = strlen( data );
	checksum = ((data[len-1]<<4)+data[len]);
	printf("#TGPS1.0 * Checksum = %i\r\n", checksum );
	return checksum;
}

int ok_checksum(char *data)
// Checks XOR Checksum (of every byte) of a sentence
// Relies on: get_checksum function
{
	char *datapointer=&data[0];
	char checksum=0;
	printf("#TGPS1.0 * Checksum is %s\r\n", data); 				// Print out the checksum in ASCII hex nibbles */
	while (*datapointer != '\0') 								// Loop through string, XOR each byte to the next byte*/
	{
		checksum ^= *datapointer;
		datapointer++;
	}
	if ( checksum != get_checksum(data) ) { 
		printf("#TGPS1.0 * Checksum is 0x%02x\r\n", checksum);  // Print out the checksum in ASCII hex nibbles
		return 0;												// Wrong
	}
	else {
		printf("#TGPS1.0 * Checksum is 0x%02x\r\n", checksum); 	// Print out the checksum in ASCII hex nibbles
																// Right
		return 1;
	}
	return 0; //should not get here
}


static double parse_latlon(char *t)
{
	double f = atof(t);
	uint16_t i = f;
	uint8_t d = i / 100;
	uint8_t m = i % 100;
	uint8_t s = (f-i) * 60;
	return d + m/60.0 + s / 3600.0;
}


static uint8_t gps_handle_line(char *l, struct gps *gps)
{ 	//<Zevv_>uit te zevven

	// $GPRMC,225446,A,4916.45,N,12311.12,W,000.5,054.7,191194,020.3,E*68
	// Recommended minimum specific GPS/Transit data 
	// 225446       Time of fix 22:54:46 UTC
	// A            Navigation receiver warning A = OK, V = warning
	// 4916.45,N    Latitude 49 deg. 16.45 min North
	// 12311.12,W   Longitude 123 deg. 11.12 min West
	// 000.5        Speed over ground, Knots
	// 054.7        Course Made Good, True
	// 191194       Date of fix  19 November 1994
	// 020.3,E      Magnetic variation 20.3 deg East
	// *68          Checksum of type XOR
	
	/* Declarations */
	
	
	return 1;
	
	
	char *t = "";
	
	if ( ok_checksum( l ) ) { 
		printf("#TGPS1.0 * Checksum Ok\r\n"); 
	} else { 
		printf("#TGPS1.0 * Checksum NOT Ok\r\n"); 
		return -1;
	};
	
	char *type = strtok(l, ",");			// Read the type of gps sentence and set char l as first occurrence for strok
											// subsequent calls to strtok must be set to NULL
	
	/* Set variables */
	gps->valid = 0;
	
	if(type == NULL) return 0;				// Check the kind of gps sentence, leave if NULL.
	
	if(strcmp(type, "$GPGGA") == 0) {
		printf("#TGPS1.0 * Got a GPGGA\r\n");
		t = strtok(NULL, ","); // Time
		if(t) {
			uint32_t v = strtol(t, NULL, 10);
			gps->H = (v / 10000);
			gps->M = (v /   100) % 100;
			gps->S = (v /     1) % 100;
		}
		t = strtok(NULL, ","); // Latitude, coord
		if(t) {
			gps->lat = parse_latlon(t);
		}
		t = strtok(NULL, ","); // Latitude, S/N
		if(t) {
			if(t[0] == 'S') gps->lat = -gps->lat;
		}		
		t = strtok(NULL, ","); // Longitude, coord
		if(t) {
			gps->lon = parse_latlon(t);
		}
		t = strtok(NULL, ","); // Longitude, W/E
		if(t) {
			if(t[0] == 'W') gps->lon = -gps->lon;
		}
		t = strtok(NULL, ","); // Fix Quality
		if(t) {
			uint32_t v = strtol(t, NULL, 10);
			gps->fixquality = v;
			if (v > 0) { gps->valid = 1; };
		}
		t = strtok(NULL, ","); // Number of Satellites IN USE
		if(t) {
			uint32_t v = strtol(t, NULL, 10);
			gps->satinuse = v;
		}
		t = strtok(NULL, ","); // Horizontal Dilution of Precision
		if(t) {
			gps->hdop = atof(t);
		}
		t = strtok(NULL, ","); // Altitude
		if(t) {
			gps->alt = atof(t);
		}
		t = strtok(NULL, ","); // Height in unit (M usually?), skip
		t = strtok(NULL, ","); // Height above WGS84 ellipsoid, skip
		t = strtok(NULL, ","); // Time since last DGPS update, (empty field usually) skip
		t = strtok(NULL, ","); // DGPS reference station ID, (empty field usually) skip
		t = strtok(NULL, ","); // Checksum, skip, we got the data already, sorry, these bits are strong.
		return 1;
	}
	
	if(strcmp(type, "$GPGSA") == 0) {
		printf("#TGPS1.0 * Got a GPGSA\r\n");	
		t = strtok(NULL, ","); // Mode, is set to always automatic (A), skip
		t = strtok(NULL, ","); // Fix Mode (1 = no fix, 2 = 2D fix, 3 = 3D fix)
		if(t) {
			uint32_t v = strtol(t, NULL, 10);
			gps->fixmode = v;
		}
		return 1;
	}
	
	if(strcmp(type, "$GPVTG") == 0) {
		printf("#TGPS1.0 * Got a GPVTG\r\n");	
		t = strtok(NULL, ","); // True Track made good, skip
		t = strtok(NULL, ","); // Magnetic track made good, skip
		t = strtok(NULL, ","); // Ground speed, knots
		t = strtok(NULL, ","); // Ground speed, km/h
		if(t) {
			uint32_t v = strtol(t, NULL, 10);
			gps->speed = v;
		}
		return 1;
	}
	else{
		printf("#TGPS1.0 * Got a %s\r\n", t);	
	}
	return 0;
}

/*
uint8_t gps_poll(struct gps *gps)
{
	static char gps_sentence[255] = "";
	static uint8_t indx = 0;
	static uint8_t data = 0;
	uint8_t r;
	if ( uart_rx_avail(1) ) {						// If data on UART1
		data = uart_rx(1);							// Get the data
		gps_sentence[indx]=data;					// Add data to char
		indx++;										// Increase index for next data
		if (data==10) { 							// if data is at the end of a sentence
			gps_sentence[indx]=0;					// Delimit char c with \0
			r = gps_handle_line(gps_sentence, gps); // Unpack data, r is true(got data)/false(got no data)
		indx = 0;									// Set index to start again
		return r;									// return true or false
		}
	}
	
	return 0;
}
*/

uint8_t gps_poll(struct gps *gps)
{
	char gps_sentence[255] = "";
	static uint8_t indx = 0;
	static uint8_t data = 0;
	
	uint8_t r;
	
	if ( uart_rx_avail(1) ) {						// If data on UART1
		data = uart_rx(1);							// Get the data
		gps_sentence[indx]=data;					// Add data to char
		indx++;										// Increase index for next data
		if (data==10) { 							// if data is at the end of a sentence
			gps_sentence[indx]=0;					// Delimit char c with \0
			strcpy( gps_sentence, gps->sentence);		// Make it so
			r = gps_handle_line(gps_sentence, gps); // OLD Unpack data, r is true(got data)/false(got no data)
			return r;								// return true or false
		}
	}
	
	return 0;
}


double gps_distance( double latitude1, double longitude1, double latitude2, double longitude2 )
{
	// One word: Haver-sine
	// e.g. http://andrew.hedges.name/experiments/haversine/
	
	// The earth is not a perfect sphere, so we use the mean radius of 6371,4445 km (~3.959mi)	
	// Since the circle of navigation will get smaller and smaller, the mean radius error in reality will become small, so we can use this for calculation
	double earth_radius = 6371.4;

	// Degree to radials 
    double deg2rad = M_PI / 180;
	
    // latitude and longitude in GPS/NMEA are in de form of degrees, and we actually don't want that 
	// alat and alon are now radians:
    double alat = deg2rad * (latitude2 - latitude1);
    double alon = deg2rad * (longitude2 - longitude1);
	
	// Haver-sine
	double a = sin(alat / 2) * sin(alat / 2) + cos(deg2rad * latitude1) * cos(deg2rad * latitude2) * sin(alon / 2) * sin(alon / 2);
    double c = 2 * asin( sqrt( a ) );
	
	// distance d is the haver-sine fraction c of the earth's radius in kilometres
    double d = earth_radius * c;
      
	return d;
}


double gps_bearing( double latitude1, double longitude1, double latitude2, double longitude2 )
{
	// atan2( SIN(lon2-lon1)*COS(lat2), COS(lat1)*SIN(lat2)-SIN(lat1)*COS(lat2)*COS(lon2-lon1) );
	// http://williams.best.vwh.net/avform.htm
	double d = fmod( atan2( sin( longitude2-longitude1 ) * cos( latitude2 ), cos( latitude1 ) * sin( latitude2 ) - sin( latitude1 ) * cos( latitude2 ) * cos( longitude2 - longitude1 ) ), 2 * M_PI );
	
	return d;
}

/*
 * End
 */